@@ -1,13 +1,11 @@
-//! Trusted publishing (via OIDC) with GitHub actions.
+//! Trusted publishing (via OIDC) with GitHub Actions and GitLab CI.
 
 use base64::Engine;
 use base64::prelude::BASE64_URL_SAFE_NO_PAD;
-use reqwest::{StatusCode, header};
+use reqwest::StatusCode;
 use reqwest_middleware::ClientWithMiddleware;
 use serde::{Deserialize, Serialize};
 use std::env;
-use std::env::VarError;
-use std::ffi::OsString;
 use std::fmt::Display;
 use thiserror::Error;
 use tracing::{debug, trace};
@@ -17,12 +15,19 @@ use uv_static::EnvVars;
 
 #[derive(Debug, Error)]
 pub enum TrustedPublishingError {
-    #[error("Environment variable {0} not set, is the `id-token: write` permission missing?")]
-    MissingEnvVar(&'static str),
-    #[error("Environment variable {0} is not valid UTF-8: `{1:?}`")]
-    InvalidEnvVar(&'static str, OsString),
     #[error(transparent)]
     Url(#[from] url::ParseError),
+    #[error("Failed to obtain OIDC token: is the `id-token: write` permission missing?")]
+    GitHubPermissions(#[source] ambient_id::Error),
+    /// A hard failure during OIDC token discovery.
+    #[error("Failed to discover OIDC token")]
+    Discovery(#[source] ambient_id::Error),
+    /// A soft failure during OIDC token discovery.
+    ///
+    /// In practice, this usually means the user attempted to force trusted
+    /// publishing outside of something like GitHub Actions or GitLab CI.
+    #[error("No OIDC token discovered: are you in a supported trusted publishing environment?")]
+    NoToken,
     #[error("Failed to fetch: `{0}`")]
     Reqwest(DisplaySafeUrl, #[source] reqwest::Error),
     #[error("Failed to fetch: `{0}`")]
@@ -38,15 +43,6 @@ pub enum TrustedPublishingError {
     InvalidOidcToken(StatusCode, String),
 }
 
-impl TrustedPublishingError {
-    fn from_var_err(env_var: &'static str, err: VarError) -> Self {
-        match err {
-            VarError::NotPresent => Self::MissingEnvVar(env_var),
-            VarError::NotUnicode(os_string) => Self::InvalidEnvVar(env_var, os_string),
-        }
-    }
-}
-
 #[derive(Deserialize)]
 #[serde(transparent)]
 pub struct TrustedPublishingToken(String);
@@ -63,12 +59,6 @@ struct Audience {
     audience: String,
 }
 
-/// The response from querying `$ACTIONS_ID_TOKEN_REQUEST_URL&audience=pypi`.
-#[derive(Deserialize)]
-struct OidcToken {
-    value: String,
-}
-
 /// The body for querying `$ACTIONS_ID_TOKEN_REQUEST_URL&audience=pypi`.
 #[derive(Serialize)]
 struct MintTokenRequest {
@@ -94,34 +84,39 @@ pub struct OidcTokenClaims {
 }
 
 /// Returns the short-lived token to use for uploading.
+///
+/// Return states:
+/// - `Ok(Some(token))`: Successfully obtained a trusted publishing token.
+/// - `Ok(None)`: Not in a supported CI environment for trusted publishing.
+/// - `Err(...)`: An error occurred while trying to obtain the token.
 pub(crate) async fn get_token(
     registry: &DisplaySafeUrl,
     client: &ClientWithMiddleware,
-) -> Result<TrustedPublishingToken, TrustedPublishingError> {
-    // If this fails, we can skip the audience request.
-    let oidc_token_request_token =
-        env::var(EnvVars::ACTIONS_ID_TOKEN_REQUEST_TOKEN).map_err(|err| {
-            TrustedPublishingError::from_var_err(EnvVars::ACTIONS_ID_TOKEN_REQUEST_TOKEN, err)
-        })?;
-
-    // Request 1: Get the audience
+) -> Result<Option<TrustedPublishingToken>, TrustedPublishingError> {
+    // Get the OIDC token's audience from the registry.
     let audience = get_audience(registry, client).await?;
 
-    // Request 2: Get the OIDC token from GitHub.
-    let oidc_token = get_oidc_token(&audience, &oidc_token_request_token, client).await?;
+    // Perform ambient OIDC token discovery.
+    // Depending on the host (GitHub Actions, GitLab CI, etc.)
+    // this may perform additional network requests.
+    let oidc_token = get_oidc_token(&audience, client).await?;
 
-    // Request 3: Get the publishing token from PyPI.
-    let publish_token = get_publish_token(registry, &oidc_token, client).await?;
+    // Exchange the OIDC token for a short-lived upload token,
+    // if OIDC token discovery succeeded.
+    if let Some(oidc_token) = oidc_token {
+        let publish_token = get_publish_token(registry, oidc_token, client).await?;
 
-    debug!("Received token, using trusted publishing");
+        // If we're on GitHub Actions, mask the exchanged token in logs.
+        #[allow(clippy::print_stdout)]
+        if env::var(EnvVars::GITHUB_ACTIONS) == Ok("true".to_string()) {
+            println!("::add-mask::{publish_token}");
+        }
 
-    // Tell GitHub Actions to mask the token in any console logs.
-    #[allow(clippy::print_stdout)]
-    if env::var(EnvVars::GITHUB_ACTIONS) == Ok("true".to_string()) {
-        println!("::add-mask::{}", &publish_token);
+        Ok(Some(publish_token))
+    } else {
+        // Not in a supported CI environment for trusted publishing.
+        Ok(None)
     }
-
-    Ok(publish_token)
 }
 
 async fn get_audience(
@@ -130,8 +125,17 @@ async fn get_audience(
 ) -> Result<String, TrustedPublishingError> {
     // `pypa/gh-action-pypi-publish` uses `netloc` (RFC 1808), which is deprecated for authority
     // (RFC 3986).
-    let audience_url =
-        DisplaySafeUrl::parse(&format!("https://{}/_/oidc/audience", registry.authority()))?;
+    // Prefer HTTPS for OIDC discovery; allow HTTP only in test builds
+    let scheme: &str = if cfg!(feature = "test") {
+        registry.scheme()
+    } else {
+        "https"
+    };
+    let audience_url = DisplaySafeUrl::parse(&format!(
+        "{}://{}/_/oidc/audience",
+        scheme,
+        registry.authority()
+    ))?;
     debug!("Querying the trusted publishing audience from {audience_url}");
     let response = client
         .get(Url::from(audience_url.clone()))
@@ -148,33 +152,24 @@ async fn get_audience(
     Ok(audience.audience)
 }
 
+/// Perform ambient OIDC token discovery.
 async fn get_oidc_token(
     audience: &str,
-    oidc_token_request_token: &str,
     client: &ClientWithMiddleware,
-) -> Result<String, TrustedPublishingError> {
-    let oidc_token_url = env::var(EnvVars::ACTIONS_ID_TOKEN_REQUEST_URL).map_err(|err| {
-        TrustedPublishingError::from_var_err(EnvVars::ACTIONS_ID_TOKEN_REQUEST_URL, err)
-    })?;
-    let mut oidc_token_url = DisplaySafeUrl::parse(&oidc_token_url)?;
-    oidc_token_url
-        .query_pairs_mut()
-        .append_pair("audience", audience);
-    debug!("Querying the trusted publishing OIDC token from {oidc_token_url}");
-    let authorization = format!("bearer {oidc_token_request_token}");
-    let response = client
-        .get(Url::from(oidc_token_url.clone()))
-        .header(header::AUTHORIZATION, authorization)
-        .send()
-        .await
-        .map_err(|err| TrustedPublishingError::ReqwestMiddleware(oidc_token_url.clone(), err))?;
-    let oidc_token: OidcToken = response
-        .error_for_status()
-        .map_err(|err| TrustedPublishingError::Reqwest(oidc_token_url.clone(), err))?
-        .json()
-        .await
-        .map_err(|err| TrustedPublishingError::Reqwest(oidc_token_url.clone(), err))?;
-    Ok(oidc_token.value)
+) -> Result<Option<ambient_id::IdToken>, TrustedPublishingError> {
+    let detector = ambient_id::Detector::new_with_client(client.clone());
+
+    match detector.detect(audience).await {
+        Ok(token) => Ok(token),
+        // Specialize the error case insufficient permissions error case,
+        // since we can offer the user a hint about fixing their permissions.
+        Err(
+            err @ ambient_id::Error::GitHubActions(
+                ambient_id::GitHubError::InsufficientPermissions(_),
+            ),
+        ) => Err(TrustedPublishingError::GitHubPermissions(err)),
+        Err(err) => Err(TrustedPublishingError::Discovery(err)),
+    }
 }
 
 /// Parse the JSON Web Token that the OIDC token is.
@@ -191,16 +186,23 @@ fn decode_oidc_token(oidc_token: &str) -> Option<OidcTokenClaims> {
 
 async fn get_publish_token(
     registry: &DisplaySafeUrl,
-    oidc_token: &str,
+    oidc_token: ambient_id::IdToken,
     client: &ClientWithMiddleware,
 ) -> Result<TrustedPublishingToken, TrustedPublishingError> {
+    // Prefer HTTPS for OIDC minting; allow HTTP only in test builds
+    let scheme: &str = if cfg!(feature = "test") {
+        registry.scheme()
+    } else {
+        "https"
+    };
     let mint_token_url = DisplaySafeUrl::parse(&format!(
-        "https://{}/_/oidc/mint-token",
+        "{}://{}/_/oidc/mint-token",
+        scheme,
         registry.authority()
     ))?;
     debug!("Querying the trusted publishing upload token from {mint_token_url}");
     let mint_token_payload = MintTokenRequest {
-        token: oidc_token.to_string(),
+        token: oidc_token.reveal().to_string(),
     };
     let response = client
         .post(Url::from(mint_token_url.clone()))
@@ -220,7 +222,7 @@ async fn get_publish_token(
         let publish_token: PublishToken = serde_json::from_slice(&body)?;
         Ok(publish_token.token)
     } else {
-        match decode_oidc_token(oidc_token) {
+        match decode_oidc_token(oidc_token.reveal()) {
             Some(claims) => {
                 // An error here means that something is misconfigured, e.g. a typo in the PyPI
                 // configuration, so we're showing the body and the JWT claims for more context, see
