@@ -12,6 +12,7 @@ use futures::TryStreamExt;
 use itertools::Itertools;
 use once_cell::sync::OnceCell;
 use owo_colors::OwoColorize;
+use reqwest_retry::policies::ExponentialBackoff;
 use reqwest_retry::{RetryError, RetryPolicy};
 use serde::Deserialize;
 use thiserror::Error;
@@ -21,7 +22,7 @@ use tokio_util::either::Either;
 use tracing::{debug, instrument};
 use url::Url;
 
-use uv_client::{BaseClient, WrappedReqwestError, is_extended_transient_error};
+use uv_client::{BaseClient, WrappedReqwestError, is_transient_network_error};
 use uv_distribution_filename::{ExtensionError, SourceDistExtension};
 use uv_extract::hash::Hasher;
 use uv_fs::{Simplified, rename_with_retry};
@@ -36,6 +37,7 @@ use crate::implementation::{
 };
 use crate::installation::PythonInstallationKey;
 use crate::managed::ManagedPythonInstallation;
+use crate::python_version::{BuildVersionError, python_build_version_from_env};
 use crate::{Interpreter, PythonRequest, PythonVersion, VersionRequest};
 
 #[derive(Error, Debug)]
@@ -97,7 +99,7 @@ pub enum Error {
     #[error("No download found for request: {}", _0.green())]
     NoDownloadFound(PythonDownloadRequest),
     #[error("A mirror was provided via `{0}`, but the URL does not match the expected format: {0}")]
-    Mirror(&'static str, &'static str),
+    Mirror(&'static str, String),
     #[error("Failed to determine the libc used on the current platform")]
     LibcDetection(#[from] platform::LibcDetectionError),
     #[error("Remote Python downloads JSON is not yet supported, please use a local path")]
@@ -110,6 +112,8 @@ pub enum Error {
         url: Box<Url>,
         python_builds_dir: PathBuf,
     },
+    #[error(transparent)]
+    BuildVersion(#[from] BuildVersionError),
 }
 
 impl Error {
@@ -142,8 +146,9 @@ impl Error {
 #[derive(Debug, PartialEq, Eq, Clone, Hash)]
 pub struct ManagedPythonDownload {
     key: PythonInstallationKey,
-    url: &'static str,
-    sha256: Option<&'static str>,
+    url: Cow<'static, str>,
+    sha256: Option<Cow<'static, str>>,
+    build: Option<&'static str>,
 }
 
 #[derive(Debug, Clone, Default, Eq, PartialEq, Hash)]
@@ -153,6 +158,7 @@ pub struct PythonDownloadRequest {
     pub(crate) arch: Option<ArchRequest>,
     pub(crate) os: Option<Os>,
     pub(crate) libc: Option<Libc>,
+    pub(crate) build: Option<String>,
 
     /// Whether to allow pre-releases or not. If not set, defaults to true if [`Self::version`] is
     /// not None, and false otherwise.
@@ -255,6 +261,7 @@ impl PythonDownloadRequest {
             arch,
             os,
             libc,
+            build: None,
             prereleases,
         }
     }
@@ -311,6 +318,12 @@ impl PythonDownloadRequest {
         self
     }
 
+    #[must_use]
+    pub fn with_build(mut self, build: String) -> Self {
+        self.build = Some(build);
+        self
+    }
+
     /// Construct a new [`PythonDownloadRequest`] from a [`PythonRequest`] if possible.
     ///
     /// Returns [`None`] if the request kind is not compatible with a download, e.g., it is
@@ -356,11 +369,25 @@ impl PythonDownloadRequest {
         Ok(self)
     }
 
+    /// Fill the build field from the environment variable relevant for the [`ImplementationName`].
+    pub fn fill_build_from_env(mut self) -> Result<Self, Error> {
+        if self.build.is_some() {
+            return Ok(self);
+        }
+        let Some(implementation) = self.implementation else {
+            return Ok(self);
+        };
+
+        self.build = python_build_version_from_env(implementation)?;
+        Ok(self)
+    }
+
     pub fn fill(mut self) -> Result<Self, Error> {
         if self.implementation.is_none() {
             self.implementation = Some(ImplementationName::CPython);
         }
         self = self.fill_platform()?;
+        self = self.fill_build_from_env()?;
         Ok(self)
     }
 
@@ -434,7 +461,31 @@ impl PythonDownloadRequest {
 
     /// Whether this request is satisfied by a Python download.
     pub fn satisfied_by_download(&self, download: &ManagedPythonDownload) -> bool {
-        self.satisfied_by_key(download.key())
+        // First check the key
+        if !self.satisfied_by_key(download.key()) {
+            return false;
+        }
+
+        // Then check the build if specified
+        if let Some(ref requested_build) = self.build {
+            let Some(download_build) = download.build() else {
+                debug!(
+                    "Skipping download `{}`: a build version was requested but is not available for this download",
+                    download
+                );
+                return false;
+            };
+
+            if download_build != requested_build {
+                debug!(
+                    "Skipping download `{}`: requested build version `{}` does not match download build version `{}`",
+                    download, requested_build, download_build
+                );
+                return false;
+            }
+        }
+
+        true
     }
 
     /// Whether this download request opts-in to pre-release Python versions.
@@ -446,6 +497,11 @@ impl PythonDownloadRequest {
         })
     }
 
+    /// Whether this download request opts-in to a debug Python version.
+    pub fn allows_debug(&self) -> bool {
+        self.version.as_ref().is_some_and(VersionRequest::is_debug)
+    }
+
     /// Whether this download request opts-in to alternative Python implementations.
     pub fn allows_alternative_implementations(&self) -> bool {
         self.implementation
@@ -753,6 +809,7 @@ struct JsonPythonDownload {
     url: String,
     sha256: Option<String>,
     variant: Option<String>,
+    build: Option<String>,
 }
 
 #[derive(Debug, Deserialize, Clone)]
@@ -767,7 +824,25 @@ pub enum DownloadResult {
     Fetched(PathBuf),
 }
 
+/// A wrapper type to display a `ManagedPythonDownload` with its build information.
+pub struct ManagedPythonDownloadWithBuild<'a>(&'a ManagedPythonDownload);
+
+impl Display for ManagedPythonDownloadWithBuild<'_> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        if let Some(build) = self.0.build {
+            write!(f, "{}+{}", self.0.key, build)
+        } else {
+            write!(f, "{}", self.0.key)
+        }
+    }
+}
+
 impl ManagedPythonDownload {
+    /// Return a display type that includes the build information.
+    pub fn to_display_with_build(&self) -> ManagedPythonDownloadWithBuild<'_> {
+        ManagedPythonDownloadWithBuild(self)
+    }
+
     /// Return the first [`ManagedPythonDownload`] matching a request, if any.
     ///
     /// If there is no stable version matching the request, a compatible pre-release version will
@@ -836,8 +911,8 @@ impl ManagedPythonDownload {
         Ok(downloads.iter())
     }
 
-    pub fn url(&self) -> &'static str {
-        self.url
+    pub fn url(&self) -> &Cow<'static, str> {
+        &self.url
     }
 
     pub fn key(&self) -> &PythonInstallationKey {
@@ -848,8 +923,12 @@ impl ManagedPythonDownload {
         self.key.os()
     }
 
-    pub fn sha256(&self) -> Option<&'static str> {
-        self.sha256
+    pub fn sha256(&self) -> Option<&Cow<'static, str>> {
+        self.sha256.as_ref()
+    }
+
+    pub fn build(&self) -> Option<&'static str> {
+        self.build
     }
 
     /// Download and extract a Python distribution, retrying on failure.
@@ -857,6 +936,7 @@ impl ManagedPythonDownload {
     pub async fn fetch_with_retry(
         &self,
         client: &BaseClient,
+        retry_policy: &ExponentialBackoff,
         installation_dir: &Path,
         scratch_dir: &Path,
         reinstall: bool,
@@ -867,7 +947,6 @@ impl ManagedPythonDownload {
         let mut total_attempts = 0;
         let mut retried_here = false;
         let start_time = SystemTime::now();
-        let retry_policy = client.retry_policy();
         loop {
             let result = self
                 .fetch(
@@ -888,7 +967,7 @@ impl ManagedPythonDownload {
                     total_attempts += err.attempts();
                     // We currently interpret e.g. "3 retries" to mean we should make 4 attempts.
                     let n_past_retries = total_attempts - 1;
-                    if is_extended_transient_error(&err) {
+                    if is_transient_network_error(&err) {
                         let retry_decision = retry_policy.should_retry(start_time, n_past_retries);
                         if let reqwest_retry::RetryDecision::Retry { execute_after } =
                             retry_decision
@@ -963,7 +1042,7 @@ impl ManagedPythonDownload {
         {
             let python_builds_dir = PathBuf::from(python_builds_dir);
             fs_err::create_dir_all(&python_builds_dir)?;
-            let hash_prefix = match self.sha256 {
+            let hash_prefix = match self.sha256.as_deref() {
                 Some(sha) => {
                     // Shorten the hash to avoid too-long-filename errors
                     &sha[..9]
@@ -1169,11 +1248,11 @@ impl ManagedPythonDownload {
         reporter: Option<&dyn Reporter>,
         direction: Direction,
     ) -> Result<(), Error> {
-        let mut hashers = self
-            .sha256
-            .into_iter()
-            .map(|_| Hasher::from(HashAlgorithm::Sha256))
-            .collect::<Vec<_>>();
+        let mut hashers = if self.sha256.is_some() {
+            vec![Hasher::from(HashAlgorithm::Sha256)]
+        } else {
+            vec![]
+        };
         let mut hasher = uv_extract::hash::HashReader::new(reader, &mut hashers);
 
         if let Some(reporter) = reporter {
@@ -1191,7 +1270,7 @@ impl ManagedPythonDownload {
         hasher.finish().await.map_err(Error::HashExhaustion)?;
 
         // Check the hash
-        if let Some(expected) = self.sha256 {
+        if let Some(expected) = self.sha256.as_deref() {
             let actual = HashDigest::from(hashers.pop().unwrap()).digest;
             if !actual.eq_ignore_ascii_case(expected) {
                 return Err(Error::HashMismatch {
@@ -1222,7 +1301,10 @@ impl ManagedPythonDownload {
                     let Some(suffix) = self.url.strip_prefix(
                         "https://github.com/astral-sh/python-build-standalone/releases/download/",
                     ) else {
-                        return Err(Error::Mirror(EnvVars::UV_PYTHON_INSTALL_MIRROR, self.url));
+                        return Err(Error::Mirror(
+                            EnvVars::UV_PYTHON_INSTALL_MIRROR,
+                            self.url.to_string(),
+                        ));
                     };
                     return Ok(Url::parse(
                         format!("{}/{}", mirror.trim_end_matches('/'), suffix).as_str(),
@@ -1234,7 +1316,10 @@ impl ManagedPythonDownload {
                 if let Some(mirror) = pypy_install_mirror {
                     let Some(suffix) = self.url.strip_prefix("https://downloads.python.org/pypy/")
                     else {
-                        return Err(Error::Mirror(EnvVars::UV_PYPY_INSTALL_MIRROR, self.url));
+                        return Err(Error::Mirror(
+                            EnvVars::UV_PYPY_INSTALL_MIRROR,
+                            self.url.to_string(),
+                        ));
                     };
                     return Ok(Url::parse(
                         format!("{}/{}", mirror.trim_end_matches('/'), suffix).as_str(),
@@ -1245,7 +1330,7 @@ impl ManagedPythonDownload {
             _ => {}
         }
 
-        Ok(Url::parse(self.url)?)
+        Ok(Url::parse(&self.url)?)
     }
 }
 
@@ -1337,9 +1422,10 @@ fn parse_json_downloads(
                 }
             };
 
-            let url = Box::leak(entry.url.into_boxed_str()) as &'static str;
-            let sha256 = entry
-                .sha256
+            let url = Cow::Owned(entry.url);
+            let sha256 = entry.sha256.map(Cow::Owned);
+            let build = entry
+                .build
                 .map(|s| Box::leak(s.into_boxed_str()) as &'static str);
 
             Some(ManagedPythonDownload {
@@ -1351,6 +1437,7 @@ fn parse_json_downloads(
                 ),
                 url,
                 sha256,
+                build,
             })
         })
         .sorted_by(|a, b| Ord::cmp(&b.key, &a.key))
@@ -1509,6 +1596,10 @@ async fn read_url(
 
 #[cfg(test)]
 mod tests {
+    use crate::implementation::LenientImplementationName;
+    use crate::installation::PythonInstallationKey;
+    use uv_platform::{Arch, Libc, Os, Platform};
+
     use super::*;
 
     /// Parse a request with all of its fields.
@@ -1637,8 +1728,8 @@ mod tests {
             request.version,
             Some(VersionRequest::from_str("3.13.0rc1").unwrap())
         );
-        assert_eq!(request.os, None,);
-        assert_eq!(request.arch, None,);
+        assert_eq!(request.os, None);
+        assert_eq!(request.arch, None);
         assert_eq!(request.libc, None);
     }
 
@@ -1722,4 +1813,90 @@ mod tests {
 
         assert!(matches!(result, Err(Error::TooManyParts(_))));
     }
+
+    /// Test that build filtering works correctly
+    #[test]
+    fn test_python_download_request_build_filtering() {
+        let request = PythonDownloadRequest::default()
+            .with_version(VersionRequest::from_str("3.12").unwrap())
+            .with_implementation(ImplementationName::CPython)
+            .with_build("20240814".to_string());
+
+        let downloads: Vec<_> = ManagedPythonDownload::iter_all(None)
+            .unwrap()
+            .filter(|d| request.satisfied_by_download(d))
+            .collect();
+
+        assert!(
+            !downloads.is_empty(),
+            "Should find at least one matching download"
+        );
+        for download in downloads {
+            assert_eq!(download.build(), Some("20240814"));
+        }
+    }
+
+    /// Test that an invalid build results in no matches
+    #[test]
+    fn test_python_download_request_invalid_build() {
+        // Create a request with a non-existent build
+        let request = PythonDownloadRequest::default()
+            .with_version(VersionRequest::from_str("3.12").unwrap())
+            .with_implementation(ImplementationName::CPython)
+            .with_build("99999999".to_string());
+
+        // Should find no matching downloads
+        let downloads: Vec<_> = ManagedPythonDownload::iter_all(None)
+            .unwrap()
+            .filter(|d| request.satisfied_by_download(d))
+            .collect();
+
+        assert_eq!(downloads.len(), 0);
+    }
+
+    /// Test build display
+    #[test]
+    fn test_managed_python_download_build_display() {
+        // Create a test download with a build
+        let key = PythonInstallationKey::new(
+            LenientImplementationName::Known(crate::implementation::ImplementationName::CPython),
+            3,
+            12,
+            0,
+            None,
+            Platform::new(
+                Os::from_str("linux").unwrap(),
+                Arch::from_str("x86_64").unwrap(),
+                Libc::from_str("gnu").unwrap(),
+            ),
+            crate::PythonVariant::default(),
+        );
+
+        let download_with_build = ManagedPythonDownload {
+            key,
+            url: Cow::Borrowed("https://example.com/python.tar.gz"),
+            sha256: Some(Cow::Borrowed("abc123")),
+            build: Some("20240101"),
+        };
+
+        // Test display with build
+        assert_eq!(
+            download_with_build.to_display_with_build().to_string(),
+            "cpython-3.12.0-linux-x86_64-gnu+20240101"
+        );
+
+        // Test download without build
+        let download_without_build = ManagedPythonDownload {
+            key: download_with_build.key.clone(),
+            url: Cow::Borrowed("https://example.com/python.tar.gz"),
+            sha256: Some(Cow::Borrowed("abc123")),
+            build: None,
+        };
+
+        // Test display without build
+        assert_eq!(
+            download_without_build.to_display_with_build().to_string(),
+            "cpython-3.12.0-linux-x86_64-gnu"
+        );
+    }
 }
