@@ -37,18 +37,19 @@ use tracing::instrument;
 use uv_cache_key::CanonicalUrl;
 use uv_client::BaseClientBuilder;
 use uv_configuration::{DependencyGroups, NoBinary, NoBuild};
-use uv_distribution_types::Requirement;
+use uv_distribution_types::{Index, Requirement};
 use uv_distribution_types::{
     IndexUrl, NameRequirementSpecification, UnresolvedRequirement,
     UnresolvedRequirementSpecification,
 };
 use uv_fs::{CWD, Simplified};
-use uv_normalize::{ExtraName, GroupName, PackageName};
+use uv_normalize::{ExtraName, PackageName, PipGroupName};
+use uv_pypi_types::PyProjectToml;
 use uv_requirements_txt::{RequirementsTxt, RequirementsTxtRequirement};
+use uv_scripts::{Pep723Error, Pep723Item, Pep723Script};
 use uv_warnings::warn_user;
-use uv_workspace::pyproject::PyProjectToml;
 
-use crate::RequirementsSource;
+use crate::{RequirementsSource, SourceTree};
 
 #[derive(Debug, Default, Clone)]
 pub struct RequirementsSpecification {
@@ -63,7 +64,7 @@ pub struct RequirementsSpecification {
     /// The `pylock.toml` file from which to extract the resolution.
     pub pylock: Option<PathBuf>,
     /// The source trees from which to extract requirements.
-    pub source_trees: Vec<PathBuf>,
+    pub source_trees: Vec<SourceTree>,
     /// The groups to use for `source_trees`
     pub groups: BTreeMap<PathBuf, DependencyGroups>,
     /// The extras used to collect requirements.
@@ -173,21 +174,150 @@ impl RequirementsSpecification {
                         ));
                     }
                 };
-                let _ = toml::from_str::<PyProjectToml>(&contents)
+                let pyproject_toml = toml::from_str::<PyProjectToml>(&contents)
                     .with_context(|| format!("Failed to parse: `{}`", path.user_display()))?;
 
                 Self {
-                    source_trees: vec![path.clone()],
+                    source_trees: vec![SourceTree::PyProjectToml(path.clone(), pyproject_toml)],
                     ..Self::default()
                 }
             }
-            RequirementsSource::SetupPy(path) | RequirementsSource::SetupCfg(path) => {
+            RequirementsSource::Pep723Script(path) => {
+                let script = match Pep723Script::read(&path).await {
+                    Ok(Some(script)) => Pep723Item::Script(script),
+                    Ok(None) => {
+                        return Err(anyhow::anyhow!(
+                            "`{}` does not contain inline script metadata",
+                            path.user_display(),
+                        ));
+                    }
+                    Err(Pep723Error::Io(err)) if err.kind() == std::io::ErrorKind::NotFound => {
+                        return Err(anyhow::anyhow!(
+                            "Failed to read `{}` (not found)",
+                            path.user_display(),
+                        ));
+                    }
+                    Err(err) => return Err(err.into()),
+                };
+
+                let metadata = script.metadata();
+
+                let requirements = metadata
+                    .dependencies
+                    .as_ref()
+                    .map(|dependencies| {
+                        dependencies
+                            .iter()
+                            .map(|dependency| {
+                                UnresolvedRequirementSpecification::from(Requirement::from(
+                                    dependency.to_owned(),
+                                ))
+                            })
+                            .collect::<Vec<UnresolvedRequirementSpecification>>()
+                    })
+                    .unwrap_or_default();
+
+                if let Some(tool_uv) = metadata.tool.as_ref().and_then(|tool| tool.uv.as_ref()) {
+                    let constraints = tool_uv
+                        .constraint_dependencies
+                        .as_ref()
+                        .map(|dependencies| {
+                            dependencies
+                                .iter()
+                                .map(|dependency| {
+                                    NameRequirementSpecification::from(Requirement::from(
+                                        dependency.to_owned(),
+                                    ))
+                                })
+                                .collect::<Vec<NameRequirementSpecification>>()
+                        })
+                        .unwrap_or_default();
+
+                    let overrides = tool_uv
+                        .override_dependencies
+                        .as_ref()
+                        .map(|dependencies| {
+                            dependencies
+                                .iter()
+                                .map(|dependency| {
+                                    UnresolvedRequirementSpecification::from(Requirement::from(
+                                        dependency.to_owned(),
+                                    ))
+                                })
+                                .collect::<Vec<UnresolvedRequirementSpecification>>()
+                        })
+                        .unwrap_or_default();
+
+                    Self {
+                        requirements,
+                        constraints,
+                        overrides,
+                        index_url: tool_uv
+                            .top_level
+                            .index_url
+                            .as_ref()
+                            .map(|index| Index::from(index.clone()).url),
+                        extra_index_urls: tool_uv
+                            .top_level
+                            .extra_index_url
+                            .as_ref()
+                            .into_iter()
+                            .flat_map(|urls| {
+                                urls.iter().map(|index| Index::from(index.clone()).url)
+                            })
+                            .collect(),
+                        no_index: tool_uv.top_level.no_index.unwrap_or_default(),
+                        find_links: tool_uv
+                            .top_level
+                            .find_links
+                            .as_ref()
+                            .into_iter()
+                            .flat_map(|urls| {
+                                urls.iter().map(|index| Index::from(index.clone()).url)
+                            })
+                            .collect(),
+                        no_binary: NoBinary::from_args(
+                            tool_uv.top_level.no_binary,
+                            tool_uv
+                                .top_level
+                                .no_binary_package
+                                .clone()
+                                .unwrap_or_default(),
+                        ),
+                        no_build: NoBuild::from_args(
+                            tool_uv.top_level.no_build,
+                            tool_uv
+                                .top_level
+                                .no_build_package
+                                .clone()
+                                .unwrap_or_default(),
+                        ),
+                        ..Self::default()
+                    }
+                } else {
+                    Self {
+                        requirements,
+                        ..Self::default()
+                    }
+                }
+            }
+            RequirementsSource::SetupPy(path) => {
                 if !path.is_file() {
                     return Err(anyhow::anyhow!("File not found: `{}`", path.user_display()));
                 }
 
                 Self {
-                    source_trees: vec![path.clone()],
+                    source_trees: vec![SourceTree::SetupPy(path.clone())],
+                    ..Self::default()
+                }
+            }
+            RequirementsSource::SetupCfg(path) => {
+                if !path.is_file() {
+                    return Err(anyhow::anyhow!("File not found: `{}`", path.user_display()));
+                }
+
+                Self {
+                    source_trees: vec![SourceTree::SetupCfg(path.clone())],
                     ..Self::default()
                 }
             }
@@ -215,7 +345,7 @@ impl RequirementsSpecification {
         requirements: &[RequirementsSource],
         constraints: &[RequirementsSource],
         overrides: &[RequirementsSource],
-        groups: BTreeMap<PathBuf, Vec<GroupName>>,
+        groups: Option<&GroupsSpecification>,
         client_builder: &BaseClientBuilder<'_>,
     ) -> Result<Self> {
         let mut spec = Self::default();
@@ -250,10 +380,13 @@ impl RequirementsSpecification {
 
         // If we have a `pylock.toml`, don't allow additional requirements, constraints, or
         // overrides.
-        if requirements
-            .iter()
-            .any(|source| matches!(source, RequirementsSource::PylockToml(..)))
-        {
+        if let Some(pylock_toml) = requirements.iter().find_map(|source| {
+            if let RequirementsSource::PylockToml(path) = source {
+                Some(path)
+            } else {
+                None
+            }
+        }) {
             if requirements
                 .iter()
                 .any(|source| !matches!(source, RequirementsSource::PylockToml(..)))
@@ -272,24 +405,55 @@ impl RequirementsSpecification {
                     "Cannot specify constraints with a `pylock.toml` file"
                 ));
             }
-            if !groups.is_empty() {
-                return Err(anyhow::anyhow!(
-                    "Cannot specify groups with a `pylock.toml` file"
-                ));
+
+            // If we have a `pylock.toml`, disallow specifying paths for groups; instead, require
+            // that all groups refer to the `pylock.toml` file.
+            if let Some(groups) = groups {
+                let mut names = Vec::new();
+                for group in &groups.groups {
+                    if group.path.is_some() {
+                        return Err(anyhow::anyhow!(
+                            "Cannot specify paths for groups with a `pylock.toml` file; all groups must refer to the `pylock.toml` file"
+                        ));
+                    }
+                    names.push(group.name.clone());
+                }
+
+                if !names.is_empty() {
+                    spec.groups.insert(
+                        pylock_toml.clone(),
+                        DependencyGroups::from_args(
+                            false,
+                            false,
+                            false,
+                            Vec::new(),
+                            Vec::new(),
+                            false,
+                            names,
+                            false,
+                        ),
+                    );
+                }
+            }
+        } else if let Some(groups) = groups {
+            // pip `--group` flags specify their own sources, which we need to process here.
+            // First, we collect all groups by their path.
+            let mut groups_by_path = BTreeMap::new();
+            for group in &groups.groups {
+                // If there's no path provided, expect a pyproject.toml in the project-dir
+                // (Which is typically the current working directory, matching pip's behaviour)
+                let pyproject_path = group
+                    .path
+                    .clone()
+                    .unwrap_or_else(|| groups.root.join("pyproject.toml"));
+                groups_by_path
+                    .entry(pyproject_path)
+                    .or_insert_with(Vec::new)
+                    .push(group.name.clone());
             }
-        }
 
-        // Resolve sources into specifications so we know their `source_tree`.
-        let mut requirement_sources = Vec::new();
-        for source in requirements {
-            let source = Self::from_source(source, client_builder).await?;
-            requirement_sources.push(source);
-        }
-
-        // pip `--group` flags specify their own sources, which we need to process here
-        if !groups.is_empty() {
             let mut group_specs = BTreeMap::new();
-            for (path, groups) in groups {
+            for (path, groups) in groups_by_path {
                 let group_spec = DependencyGroups::from_args(
                     false,
                     false,
@@ -305,6 +469,13 @@ impl RequirementsSpecification {
             spec.groups = group_specs;
         }
 
+        // Resolve sources into specifications so we know their `source_tree`.
+        let mut requirement_sources = Vec::new();
+        for source in requirements {
+            let source = Self::from_source(source, client_builder).await?;
+            requirement_sources.push(source);
+        }
+
         // Read all requirements, and keep track of all requirements _and_ constraints.
         // A `requirements.txt` can contain a `-c constraints.txt` directive within it, so reading
         // a requirements file can also add constraints.
@@ -426,7 +597,7 @@ impl RequirementsSpecification {
         requirements: &[RequirementsSource],
         client_builder: &BaseClientBuilder<'_>,
     ) -> Result<Self> {
-        Self::from_sources(requirements, &[], &[], BTreeMap::default(), client_builder).await
+        Self::from_sources(requirements, &[], &[], None, client_builder).await
     }
 
     /// Initialize a [`RequirementsSpecification`] from a list of [`Requirement`].
@@ -485,3 +656,12 @@ impl RequirementsSpecification {
         self.requirements.is_empty() && self.source_trees.is_empty() && self.overrides.is_empty()
     }
 }
+
+#[derive(Debug, Default, Clone)]
+pub struct GroupsSpecification {
+    /// The path to the project root, relative to which the default `pyproject.toml` file is
+    /// located.
+    pub root: PathBuf,
+    /// The enabled groups.
+    pub groups: Vec<PipGroupName>,
+}
